import torch
import random
import glob
import json
from load_pretrained_cls import load_pretrained_cls
from utils import load_checkpoint, dict2clsattr, fix_all_seed
from os.path import abspath, exists, join
from sampling import Sampler
from argparse import ArgumentParser
from torch.backends import cudnn


def load_gen_cls(cfgs, device):
    module = __import__('models.{architecture}'.format(architecture=cfgs.architecture), fromlist=['something'])

    Gen = module.Generator(cfgs.z_dim, cfgs.shared_dim, cfgs.img_size, cfgs.g_conv_dim, cfgs.g_spectral_norm, cfgs.attention, cfgs.attention_after_nth_gen_block, cfgs.activation_fn, cfgs.conditional_strategy, cfgs.num_classes, cfgs.g_init, cfgs.G_depth).to(device)
    Cls = load_pretrained_cls(cfgs.dataset_name, cfgs.conditional_strategy).to(device)

    when = "best"

    if not exists(abspath(cfgs.checkpoint_dir)):
        raise NotADirectoryError
    
    g_checkpoint_dir = glob.glob(join(cfgs.checkpoint_dir,"model=G-best-weights-step*.pth".format(when=when)))[0]
    Gen = load_checkpoint(Gen, g_checkpoint_dir)
    c_checkpoint_dir = glob.glob(join(cfgs.checkpoint_dir,f"model=C-best-weights-step*.pth"))[0]
    Cls = load_checkpoint(Cls, c_checkpoint_dir)

    return Gen, Cls


def sample(cfgs, Gen, Cls, device):
    session_dir = join(cfgs.result_dir, cfgs.session_name)
    sampler = Sampler(
        session_dir=session_dir,
        Gen=Gen,
        Cls=Cls,
        conditional_strategy=cfgs.conditional_strategy,
        dataset_name=cfgs.dataset_name,
        z_dim=cfgs.z_dim,
        gammas=cfgs.gammas,
        temperature_logits=cfgs.temperature_logits,
        temperature_validity=cfgs.temperature_validity,
        device=device
    )

    sampler.mh(
        batch_size=cfgs.batch_size,
        intersection_indices=cfgs.intersection_indices,
        difference_indices=cfgs.difference_indices,
        n_steps=cfgs.n_steps,
        save_img=True,
        save_every=cfgs.save_every
    )


def main():
    parser = ArgumentParser(add_help=False)
    parser.add_argument('--config_path', type=str, default=None)
    parser.add_argument('--checkpoint_dir', type=str, default=None)
    parser.add_argument('--seed', type=int, default=-1, help='seed for generating random numbers')

    # sampling
    parser.add_argument('--batch_size', type=int, default=64)
    parser.add_argument('--result_dir', type=str, default='./results', help='results directory')
    parser.add_argument('--session_name', type=str, default=None, help='name of directory for saving images')
    parser.add_argument('--n_steps', type=int, default=200, help='MCMC steps for Metropolis-hastings sampling')
    parser.add_argument('--intersection_indices', type=int, nargs='+', default=[0, 1, 2])
    parser.add_argument('--difference_indices', type=int, nargs='+', default=[])
    parser.add_argument('--temperature_logits', type=float, default=1.0)
    parser.add_argument('--temperature_validity', type=float, default=1.0)
    parser.add_argument('--gammas', type=float, nargs='+', default=[1.0, 1.0, 1.0])
    parser.add_argument('--save_every', type=int, default=10)

    args = parser.parse_args()

    if args.config_path is not None:
        with open(args.config_path) as f:
            model_configs = json.load(f)
        cfgs = vars(args)
        cfgs = dict2clsattr(cfgs, model_configs)
    else:
        raise NotImplementedError

    if cfgs.seed == -1:
        cfgs.seed = random.randint(1,4096)
        cudnn.benchmark, cudnn.deterministic = True, False
    else:
        cudnn.benchmark, cudnn.deterministic = False, True

    fix_all_seed(cfgs.seed)
    device = torch.cuda.current_device()

    Gen, Cls = load_gen_cls(cfgs, device)
    sample(cfgs, Gen, Cls, device)


if __name__ == '__main__':
    main()
